/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2008 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person 
 *  obtaining a copy of this software and associated documentation 
 *  files (the "Software"), to deal in the Software without 
 *  restriction, including without limitation the rights to use, copy, 
 *  modify, merge, publish, distribute, sublicense, and/or sell copies 
 *  of the Software, and to permit persons to whom the Software is 
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind, 
 *  express or implied, including but not limited to the warranties of 
 *  merchantability, fitness for a particular purpose and 
 *  noninfringement. In no event shall the authors or copyright 
 *  holders be liable for any claim, damages or other liability, 
 *  whether in an action of contract, tort or otherwise, arising from, 
 *  out of or in connection with the software or the use or other 
 *  dealings in the software.
 */

#ifndef CROW_VALUE_H
#define CROW_VALUE_H

#include "util.h"

namespace Crow {

class CAny : public Object {

public:
	static Any createInt(int data);
	static Any createUnsigned(unsigned data);
	static Any createBool(bool data);
	static Any createFloat(float data);
	static Any createDouble(double data);
	static Any createString(const Glib::ustring & data);
	static Any createEnum(TypeId type, int data);
	static Any createEnum(const std::string & tname, int data);
	static Any createFlags(TypeId type, Bits data);
	static Any createFlags(const std::string & tname, Bits data);
	static Any createObject(TypeId type, PGlibObject data);
	static Any createObject(PGlibObject data);
	static Any createObject();
	static Any createVector(const AnyVector & data);
	static Any createPoint(const Point & data);
	static Any createEmitter(const Emitter & emitter);

	template <class T> static Any create(TypeId type, const T & data);
	template <class T> static Any create(const std::string & type, const T & data);

	static Any createFromGValue(TypeId type, const Glib::ValueBase & data);
	static Any createFromString(TypeId type, const Glib::ustring & data);

	TypeId getType() const { return type; }

	int getInt() const;
	unsigned getUnsigned() const;
	bool getBool() const;
	float getFloat() const;
	double getDouble() const;
	const Glib::ustring * getString() const;
	int getEnum() const;
	Bits getFlags() const;
	PGlibObject getObject() const;
	const AnyVector * getVector() const;
	const Point * getPoint() const;
	const Emitter * getEmitter() const;

	template <class T> Glib::RefPtr<T> getObject() const;
	template <class T> const T * get() const;

	virtual Glib::ustring toString() const = 0;
	virtual Glib::ValueBase toGValue() const = 0;

protected:
	CAny(TypeId tid);
	CAny(const std::string & tname);

	template <class A, class B> static bool equal(A * a, B * b);

private:
	TypeId type;

};

template <class T> class TAny : public CAny {

friend class CAny;

public:
	bool equals(PObject v) const { return CAny::equal(this, GetPtr(Any::cast_dynamic(v))); }
	Glib::ustring toString() const { return ToString<T>(value); }
	Glib::ValueBase toGValue() const { return MakeGValue<T>(value); }

protected:
	TAny(TypeId type, const T & v) :
		CAny(type), value(v)
	{ }

	const T * getValue() const { return &value; }

private:
	T value;

};

template <class A, class B> bool CAny::equal(A * a, B * b) {
	if(a==b)
		return true;
	if(!a || !b)
		return false;
	if(a->getType() != b->getType())
		return false;
	A * ab= dynamic_cast<A *>(b);
	if(!ab)
		return false;
	return *a->getValue() == *ab->getValue();
}

template <class T> Any CAny::create(TypeId type, const T & data) {
	return NewRefPtr(new TAny<T>(type, data));
}

template <class T> Any CAny::create(const std::string & type, const T & data) {
	TypeId tid= FindType(type);
	return CAny::create<T>(tid, data);
}

template <class T> const T * CAny::get() const {
	const TAny<T> * tany= dynamic_cast<const TAny<T> *>(this);
	Check(tany);
	return tany->getValue();
}

template <class T> Glib::RefPtr<T> CAny::getObject() const {
	PGlibObject object= *get<PGlibObject>();
	return Glib::RefPtr<T>::cast_dynamic(object);
}

template <> inline Glib::ustring ToString<AnyVector>(const AnyVector & value) { return ""; }

class CVector : public TAny<AnyVector> {

friend class CAny;

public:
	bool equals(PObject v) const;

protected:
	CVector(const AnyVector & v);

};

class CEnum : public TAny<int> {

friend class CAny;

public:
	Glib::ustring toString() const;

protected:
	CEnum(TypeId tid, int v);

};

class CFlags : public TAny<Bits> {

friend class CAny;

public:
	Glib::ustring toString() const;

protected:
	CFlags(TypeId tid, Bits v);

};

}

#endif
